#lang racket

(require "mascheme.rkt")
(require "test.rkt")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;; examples from paper ;;;;;;;;;

;; Basic utility functions

(define (run-browser)
  (get-url "http://test.url")
  (access-disk))

(define (get-url url)
  (write 'Url))

(define (access-disk)
  (write 'Disk))

(define (check-cache)
  (access-disk))

;; caching aspect

(define (((cache proceed) . ctx) . args)
  (write 'Cache)
  (check-cache)
  (apply proceed args))

;; quota aspect

(define (((quota proceed) . ctx) . args)
  (write 'Quota)
  (apply proceed args))

;; Figure 5a: tower
(let ((l0 (new-membrane))
      (l1 (new-membrane))
      (l2 (new-membrane)))
  (advise l1 l0)
  (advise l2 l1)
  (register (call get-url) cache l1)
  (register (call access-disk) quota l2)
  (test/str 
   (deploy-fluid l0
                 (run-browser))
   "CacheQuotaDiskUrlDisk"))

;; Figure 5b: tree
(let ((l0 (new-membrane))
      (l1 (new-membrane))
      (l2 (new-membrane))
      (l3 (new-membrane)))
  (advise l1 l0)
  (advise l2 l1)
  (advise l3 l0)
  (register (call get-url) cache l1)
  (register (call access-disk) quota l2)
  (register (call access-disk) quota l3)
  (test/str 
   (deploy-fluid l0
                 (run-browser))
   "CacheQuotaDiskUrlQuotaDisk"))

;; Figure 5c: graph
(let ((l0 (new-membrane))
      (l1 (new-membrane))
      (l2 (new-membrane)))
  (advise l1 l0)
  (advise l2 l0)
  (advise l2 l1)
  (register (call get-url) cache l1)
  (register (call access-disk) quota l2)
  (test/str
   (deploy-fluid l0
                 (run-browser))
   "CacheQuotaDiskUrlQuotaDisk"))

;; Fig8: crosscutting membranes
(define logger
  (lambda (proceed) 
    (lambda ctx 
      (lambda args 
        (write 'Log)
        (apply proceed args)))))

(let ((m-admin-user (new-membrane))
      (m-user-guest (new-membrane))
      (m-cache      (new-membrane))
      (m-log        (new-membrane)))
  
  (define user-browser
    (deploy-static m-admin-user
            (deploy-static m-user-guest
                    (lambda () 
                      (write 'UserB)
                      (get-url "http://test.url")))))
  
  (define admin-browser
    (deploy-static m-admin-user
            (lambda () 
              (write 'AdminB)
              (get-url "http://test.url"))))
  
  (define guest-browser
    (deploy-static m-user-guest
            (lambda () 
              (write 'GuestB)
              (get-url "http://test.url"))))
  
  (register (call get-url) cache  m-cache)
  (register (call get-url) logger m-log)
  
  (advise m-cache m-admin-user)
  (advise m-log   m-user-guest)
  
  (test/str
   (begin (user-browser)
          (write '-)
          (admin-browser)
          (write '-)
          (guest-browser))
   "UserBLogCacheDiskUrl-AdminBCacheDiskUrl-GuestBLogUrl"))
  

;; examples of programmable advised membrane: 
;; a membrane can define a pointcut that filters the join points that are
;; actually emitted (ie. "a la Open Modules")

;; opaque membrane
(let ((l0 (new-membrane (λ args #f)))
      (l1 (new-membrane)))
  (advise l1 l0)
  (register (call get-url) cache l1)
  (test/str
   (deploy-fluid l0
                 (run-browser))
   "UrlDisk"))

;; translucent membrane
(define trace
  (lambda (proceed) 
    (lambda ctx 
      (lambda args 
        (write 'Trace)
        (apply proceed args)))))

; expose only no-arg jps
(let ((l0 (new-membrane 
           ; home-made pointcut
           (λ (jp- jp jp+) (pc-pred (= 0 (length (args jp- jp jp+)))))))
      (l1 (new-membrane)))
  (advise l1 l0)
  (register (|| (call get-url) (call access-disk)) trace l1)
  (test/str
   (deploy-fluid l0
                 (run-browser))
   "UrlTraceDisk"))

(define (string-prefix? s p)
  (equal? (substring s 0 (string-length p))
          p))

;; expose only calls to get-url for non-https requests
(let ((l0 (new-membrane 
           (&& (call get-url)
               (args? (λ (args) (and (= 1 (length args))
                                     (string? (car args))
                                     (string-prefix? (car args) "http://")))))))
      (l1 (new-membrane)))
  (advise l1 l0)
  (register (|| (call get-url) (call access-disk)) trace l1)
  (test/str
   (deploy-fluid l0
                 (run-browser))
   "TraceUrlDisk"))

;; expose only calls to get-url for https requests
(let ((l0 (new-membrane 
           (&& (call get-url)
               (args? (λ (args) (and (= 1 (length args))
                                     (string? (car args))
                                     (string-prefix? (car args) "https://")))))))
      (l1 (new-membrane)))
  (advise l1 l0)
  (register (|| (call get-url) (call access-disk)) trace l1)
  (test/str
   (deploy-fluid l0
                 (run-browser))
   "UrlDisk"))


