#lang scheme

;;; MAScheme -- AspectScheme with Membranes
;;;  by Éric Tanter
;;; Based on: 
;;; AspectScheme
;;; Copyright (c) 2005-2008 by Christopher Dutchyn (cdutchyn@cs.usask.ca); all rights reserved.
;;;
;;; note: all comments only related to AspectScheme have been cut/shortened.

(require (only-in swindle/base let let* letrec))

(provide (except-out (all-from-out scheme) lambda #%app)
         (rename-out [app/weave #%app] 
                     [lambda/membs lambda])
         ;; membrane creation and advising
         new-membrane
         advise
         ;; aspect registration
         register
         ;; membrane deployment
         deploy-fluid
         deploy-static
         deploy-top
         ;; pointcuts & co
         && || !
         top? top below above bottom bottom?
         target args some-args args?
         call? exec? call exec
         cflowtop cflowbelow cflowbottom cflowabove
         cflow within
         let let* letrec
         pc-pred
         
         ;; debug
         membrane-advisings membrane-aspects
         )

;; aspect structure
(define-struct aspect (pc adv))

(define-struct          jp (target args) #:mutable) 
(define-struct (call-jp jp)(           ))
(define-struct (exec-jp jp)(           ))

;; membrane structure
(define-struct membrane (advisings 
                         aspects
                         out-filter) #:mutable)

(define (new-membrane [out-filter (λ args '())])
  (make-membrane '() '() out-filter))

(define (advise ma mb)
  (set-membrane-advisings! mb (cons ma (membrane-advisings mb))))

(define (register pc adv m)
  (set-membrane-aspects! m (cons (make-aspect pc adv) (membrane-aspects m))))

;; from AspectScheme (see doc there)
(define-syntax fluid-let-parameter
  (syntax-rules ()
    [(_ ([p v]) e ...)
     (let ([y v])
       (let ([(swap) (let ([t (p)])
                       (p y)
                       (set! y t))])
         (dynamic-wind swap
                       (lambda () e ...)
                       swap)))]))

;; statically and dynamically scoped membranes
;;;
(define dynamic-membranes (make-parameter '()))
(define  static-membranes (make-parameter '()))

(define-syntaxes (deploy-fluid deploy-static)
  (let ([wrap (lambda (param)
                 (lambda (stx)
                   (syntax-case stx ()
                     [(_ m body ...)
                      (quasisyntax/loc stx 
                        (fluid-let-parameter ([#,param (cons m (#,param))])
                          body ...))])))])
    (values (wrap #`dynamic-membranes)   ;dynamically-scoped
            (wrap #`static-membranes)))) ;lexically-scoped

;; lexically-scoped membranes
(define-syntax lambda/membs 
  (lambda (stx)
    (syntax-case stx ()
      [(_ params body ...)
       (syntax/loc stx (let ([membs (static-membranes)])
                         (lambda params
                           (fluid-let-parameter ([static-membranes membs])
                             body ...))))])))
;; top-level membranes
(define toplevel-membranes (make-parameter '()))

(define (deploy-top m)
  (toplevel-membranes (cons m (toplevel-membranes))))

;; weaver
;; current membranes -- in descending order of application!
(define (current-membranes)
  (append (dynamic-membranes)
          (static-membranes)
          (toplevel-membranes)))

;; join points implemented as continuation marks
(define-struct jp-mark (id))
(define jp-mark-tag (make-jp-mark 'joinpoint))

(define (jp-context) 
  (continuation-mark-set->list
   (current-continuation-marks) 
   jp-mark-tag))

(define-syntax with-joinpoint
  (syntax-rules ()
    [(_ jp body ...)
     ((lambda (x) x)
      (with-continuation-mark jp-mark-tag jp
        (begin body ...)))]))

;; replacement for #%app
(define-syntax app/weave
  (lambda (stx)
    (syntax-case stx ()
      [(_ f a ...)
       (syntax/loc stx (app/weave/rt f a ...))])))

;; ancillary
(define (filter-void l)
  (if (empty? l) l
      (if (void? (car l))
          (filter-void (cdr l))
          (cons (car l) (filter-void (cdr l))))))


(define (app/weave/rt fun-val . arg-vals) 
  (if (primitive? fun-val)
      (apply fun-val arg-vals)
      (let ([jp (make-call-jp fun-val arg-vals)]
            [jp+ (jp-context)])
        (let ((current-dyn (dynamic-membranes))) ;; capture dyn membranes
          (with-joinpoint jp              
            (let ((base-eval (lambda arg-vals
                               (parameterize ((dynamic-membranes current-dyn)) ;; restore on last proceed
                                 (with-joinpoint (make-exec-jp fun-val arg-vals)
                                                 (apply fun-val arg-vals))))))
              
              ;; for each current membrane, check if jp is propagated.
              ;; if so, weave all aspects of advising membranes
              (apply (foldr 
                      (lambda (cm fun)
                        (if (eval-prim (λ () ((membrane-out-filter cm) '() jp jp+)))
                            (foldr (lambda (m f)
                                     (lambda arg-vals
                                       (parameterize ((dynamic-membranes (list m)))
                                         (apply (weave f '() jp jp+ (membrane-aspects m))
                                                arg-vals))))
                                   fun
                                   (membrane-advisings cm))
                            fun))
                      base-eval
                      (current-membranes))
                     arg-vals)))))))

;; eval thunk deactivating context membranes (dynamic and top-level)
;; used to evaluate custom pointcuts given to define membrane translucence
(define (eval-prim thunk)
  (parameterize ((dynamic-membranes '())
                 (toplevel-membranes '()))
    (thunk)))
              
;; weave all aspects within a membrane
(define (weave fun-val jp- jp jp+ aspects)
  (foldr (lambda (aspect fun)
           (cond
             [(with-joinpoint (make-call-jp (aspect-pc aspect) (list jp- jp jp+))
                              ((aspect-pc aspect) jp- jp jp+))                  
              => (lambda (ctxt-vals)
                   (lambda args
                   (with-joinpoint (make-call-jp (aspect-adv aspect) ctxt-vals)
                                   (apply (apply ((aspect-adv aspect) fun) ctxt-vals) args))))] 
             [else fun]))                                                                       
         fun-val
         aspects))

;; pointcuts
(define (pc-pred pred)
  (and pred '()))

;; pointcuts -- strict combinators
;; NB. This PLT Scheme module does not export app/weave for #%app
;; until the end of the module, so these definitions are not woven
(define ((&& . pcs) jp- jp jp+)
  (let loop ([pcs  pcs]
             [res  '()])
    (if (null? pcs)
        res
        (let ([r ((car pcs) jp- jp jp+)])
          (and r
               (loop (cdr pcs) (append res r)))))))

(define ((|| . pcs) jp- jp jp+)
  (let loop ([pcs pcs])
    (and (not (null? pcs))
         (or ((car pcs) jp- jp jp+)
             (loop (cdr pcs))))))

(define ((! pc) jp- jp jp+)
  (and (not (pc jp- jp jp+))
       '()))

;; pointcuts -- structural
(define (top? jp- jp jp+)
  (and (null? jp+)
       '()))

(define ((below pc) jp- jp jp+)
  (and (not (null? jp+))
       (pc (cons jp jp-) (car jp+) (cdr jp+))))

(define ((above pc) jp- jp jp+)
  (and (not (null? jp-))
       (pc (cdr jp-) (car jp-) (cons jp jp+))))

(define (bottom? jp- jp jp+)
  (and (null? jp-)
       '()))

;; pointcuts -- `binding'
(define (target jp- jp jp+)
  (list (jp-target jp)))

(define (args jp- jp jp+)
  (jp-args jp))

(define ((some-args as) jp- jp jp+)
  (foldl (lambda (a v l)
           (if a
               (cons v l)
               l))
         '()
         as
         (jp-args jp)))


;; pointcuts -- fundamental
(define ((true-pc) jp- jp jp+)
  '())

(define ((this-jp? p?) jp- jp jp+)
  (and (p? jp)
       '()))

(define ((args? p?) jp- jp jp+)
  (and (p? (jp-args jp))
       '()))

(define-values (call? exec?)   
  (values (this-jp? call-jp?)
          (this-jp? exec-jp?)))

(define-values (call exec)      
  (let* ([(target= f) (this-jp? (lambda (jp)
                                      (eqv? f (jp-target jp))))]
         [((this-target? p) f) (&& (target= f)
                                  p)])
    (values (this-target? call?)
            (this-target? exec?))))

(define (display-jps jp- jp jp+)
  (for-each (lambda (jp) (display `(< ,jp ,(jp-args jp)))) jp-)
  (display `(= ,jp ,(jp-args jp)))
  (for-each (lambda (jp) (display `(> ,jp ,(jp-args jp)))) jp+)
  (newline))

;; pointcuts - higher-order recursive
(define (((cflow-walk pc) step end) jp- jp jp+)
  ((|| pc
       (&& (! end)
           (step ((cflow-walk pc) step end)))) jp- jp jp+))

(define (top pc)
  (&& pc
      (! (cflowbelow pc))))

(define (bottom pc)
  (&& pc
      (! (cflowabove pc))))

;; pointcuts - higher-order points-free
(define (cflowtop pc)
  (cflowbelow (top pc)))

(define (cflowbelow pc)
  (below ((cflow-walk pc) below top?)))

(define (cflowabove pc)
  (above ((cflow-walk pc) above bottom)))

(define (cflowbottom pc)
  (cflowbelow (bottom pc)))

;; pointcuts - compatibility
(define (cflow pc)
  ((cflow-walk pc) below top?))

(define (within f)
  (cflowbelow (&& (exec f)
                  (! (cflowabove call?)))))

